/*
    SPDX-FileCopyrightText: 2011 Rafał Kułaga <rl.kulaga@gmail.com>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "pwizchartconfig.h"

PWizChartConfigUI::PWizChartConfigUI(QWidget *parent) : QFrame(parent)
{
    setupUi(this);
}
