/*
    SPDX-FileCopyrightText: 2012 Jasem Mutlaq <mutlaqja@ikarustech.com>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "astrometryparser.h"

namespace Ekos
{
// At least a dtor is needed for the correct mocing
AstrometryParser::AstrometryParser()
{
}
}
