/*  KStars class tests
    SPDX-FileCopyrightText: 2020 Eric Dejouhanet <eric.dejouhanet@gmail.com>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "testfov.h"

TestFOV::TestFOV(QObject *parent) : QObject(parent)
{
}

void TestFOV::initTestCase()
{
}

void TestFOV::cleanupTestCase()
{
}

void TestFOV::init()
{
}

void TestFOV::cleanup()
{
}

void TestFOV::testFOVCreation()
{
    QSKIP("Not implemented yet.");
}

QTEST_GUILESS_MAIN(TestFOV)
