/*
    SPDX-FileCopyrightText: 2005 Jason Harris <kstars@30doradus.org>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "coordinategrid.h"

CoordinateGrid::CoordinateGrid(SkyComposite *parent, const QString &name) : NoPrecessIndex(parent, name)
{
}
