/*  KStars class tests
    SPDX-FileCopyrightText: 2020 Eric Dejouhanet <eric.dejouhanet@gmail.com>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "testgeolocation.h"

TestGeolocation::TestGeolocation(QObject *parent) : QObject(parent)
{
}

void TestGeolocation::initTestCase()
{
}

void TestGeolocation::cleanupTestCase()
{
}

void TestGeolocation::init()
{
}

void TestGeolocation::cleanup()
{
}

void TestGeolocation::testGeolocation_data()
{
}

void TestGeolocation::testGeolocation()
{
    QSKIP("Not implemented yet.");
}

void TestGeolocation::testParseCityDatabase()
{
    QSKIP("Not implemented yet.");
}

void TestGeolocation::testCoordinates()
{
    QSKIP("Not implemented yet.");
}

QTEST_GUILESS_MAIN(TestGeolocation)
