/*
    SPDX-FileCopyrightText: 2005 Jason Harris and Jasem Mutlaq <kstars@30doradus.org>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "clicklabel.h"

ClickLabel::ClickLabel(QWidget *parent, const char *name) : QLabel(parent)
{
    setObjectName(name);
}
